# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax.numpy as jnp
from blackjax.types import PyTree, PRNGKey

#: JAX Array
Array = jnp.ndarray

#: JAX PyTrees
PyTree = PyTree

#: JAX PRNGKey
PRNGKey = PRNGKey